/*
 * gkacct.cxx
 *
 * Accounting modules for GNU Gatekeeper. Provides generic
 * support for accounting to the gatekeeper.
 *
 * Copyright (c) 2003, Quarcom FHU, Michal Zygmuntowicz
 *
 * This work is published under the GNU Public License (GPL)
 * see file COPYING for details.
 * We also explicitely grant the right to link this code
 * with the OpenH323 library.
 *
 * $Log: gkacct.cxx,v $
 * Revision 1.1.2.1  2003/06/19 15:36:04  zvision
 * Initial generic accounting support for GNU GK.
 *
 */
#include "gk_const.h"
#include "h323util.h"
#include "Toolkit.h"
#include "RasTbl.h"
#include "gkacct.h"

/// Name of the config file section for accounting configuration
const char* GkAcctSectionName = "Gatekeeper::Acct";

PString GkAcctLogger::AsString(
	const PTime& tm
	)
{
	struct tm _tm;
	struct tm* tmptr = &_tm;
	time_t t;
	
	if( (time(&t) != (time_t)(-1))
#ifndef WIN32
		&& (localtime_r(&t,tmptr) == tmptr) )
#else
		&& ((tmptr = localtime(&t)) != NULL) )
#endif
	{
		char buf[10];
		
		buf[0] = 0;
		if( strftime(buf,sizeof(buf),"%Z",tmptr) < 10 )
		{
			buf[9] = 0;
			const PString tzname(buf);
			if( !tzname.IsEmpty() )
			{
				PString s = tm.AsString( "hh:mm:ss.uuu @@@ www MMM d yyyy" );
				s.Replace("@@@",tzname);
				return s;
			}
		}
	}
	
	return tm.AsString( "hh:mm:ss.uuu z www MMM d yyyy" );
}

GkAcctLogger::GkAcctLogger(
	PConfig& cfg, 
	const PString& moduleName
	) 
	: 
	controlFlag(Sufficient),
	defaultStatus(Fail),
	eventMask(AcctAll),
	name(moduleName)
{
	const PStringArray control( 
		cfg.GetString( GkAcctSectionName, name, "" ).Tokenise(";,")
		);

	if( control.GetSize() < 1 )
		PTRACE(1,"GKACCT\tEmpty config entry for module "<<moduleName);
	else if( name *= "default" )
	{
		controlFlag = Sufficient,
		defaultStatus = Toolkit::AsBool(control[0]) ? Ok : Fail;
	}
	else if (control[0] *= "optional")
	{
		controlFlag = Optional;
		defaultStatus = Next;
	}
	else if (control[0] *= "required")
	{
		controlFlag = Required;
		defaultStatus = Fail;
	}
	
	if( control.GetSize() > 1 )
		eventMask = ReadEventMask(control);
	
	PTRACE(1,"GKACCT\tCreated module "<<name<<" with event mask "
		<<PString(PString::Unsigned,(long)eventMask,16)
		);
}

GkAcctLogger::~GkAcctLogger()
{
	PTRACE(1,"GKACCT\tDestroyed module "<<name);
}

int GkAcctLogger::ReadEventMask(
	const PStringArray& tokens
	) const
{
	int mask = 0;
	
	for( PINDEX i = 1; i < tokens.GetSize(); i++ )
	{
		const PString& token = tokens[i];
		if( token *= "start" )
			mask |= AcctStart;
		else if( token *= "stop" )
			mask |= AcctStop;
		else if( token *= "update" )
			mask |= AcctUpdate;
		else if( token *= "on" )
			mask |= AcctOn;
		else if( token *= "off" )
			mask |= AcctOff;
	}
	
	return mask;
}

GkAcctFactoryBase::GkAcctFactoryArray* GkAcctFactoryBase::acctLoggerFactories = NULL;

GkAcctFactoryBase::GkAcctFactoryBase(
	const PString& moduleName
	) 
	: 
	name(moduleName)
{
	static GkAcctFactoryArray factories;
	acctLoggerFactories = &factories;
	
	PINDEX i;
	
	for( i = 0; i < acctLoggerFactories->GetSize(); i++ )
	{
		GkAcctFactoryBase *const factory = acctLoggerFactories->GetAt(i);
		if( factory && ((factory == this) || (factory->GetName() == moduleName)) )
		{
			PTRACE(1,"GKACCT\tGlobal list of modules already contains"
				" a factory for "<<name
				);
			break;
		}
	}
	
	if( i >= acctLoggerFactories->GetSize() )
	{
		acctLoggerFactories->SetAt(i,this);
		PTRACE(3,"GKACCT\tRegistered factory for "<<name);
	}
}

GkAcctFactoryBase::~GkAcctFactoryBase()
{
}

PObject::Comparison GkAcctFactoryBase::Compare( 
	const PObject& obj
	) const
{
	if( !obj.IsDescendant(GkAcctFactoryBase::Class()) )
		return LessThan;
	else
		return name.Compare( ((const GkAcctFactoryBase&)obj).name );
}

GkAcctFactoryBase* GkAcctFactoryBase::FindFactoryFor(
	const PString& moduleName
	)
{
	if( acctLoggerFactories != NULL )
		for( PINDEX i = 0; i < acctLoggerFactories->GetSize(); i++ )
		{
			GkAcctFactoryBase *const factory = acctLoggerFactories->GetAt(i);
			if( factory && (factory->GetName() == moduleName) )
				return factory;
		}
	return NULL;
}

GkAcctLoggers::GkAcctLoggers(
	PConfig& cfg
	)
{
	const PStringArray acctLoggerNames( 
		cfg.GetKeys(GkAcctSectionName) 
		);

	for( PINDEX i = 0; i < acctLoggerNames.GetSize(); i++ ) 
	{
		GkAcctFactoryBase* acctFactory 
			= GkAcctFactoryBase::FindFactoryFor(acctLoggerNames[i]);

		if( acctFactory != NULL )
			Append(acctFactory->CreateAcctLogger(cfg));
		else
			PTRACE(1,"GKACCT\tUnknown accounting logger module name "
				<<acctLoggerNames[i]<< " - not created"
				);
	}
}

GkAcctLoggers::~GkAcctLoggers()
{
}

BOOL GkAcctLoggers::LogAcctEvent( 
	GkAcctLogger::AcctEvent evt, /// accounting event to log
	callptr& call /// additional data for the event
	)
{
	BOOL finalResult = TRUE;
	
	for( PINDEX i = 0; i < GetSize(); i++ )
	{
		GkAcctLogger* logger = (GkAcctLogger*)GetAt(i);
		if( logger == NULL )
		{
			PTRACE(1,"GKACCT\tNULL logger on the list");
			continue;
		}
		
		if( (logger->GetEventMask() & evt) == 0 )
			continue;
			
		GkAcctLogger::Status status = logger->LogAcctEvent( evt, call );
		switch( status )
		{
		case GkAcctLogger::Ok:
			if( logger->GetControlFlag() == GkAcctLogger::Sufficient )
				return TRUE;
			break;
			
		case GkAcctLogger::Next:
			if( logger->GetControlFlag() != GkAcctLogger::Optional )
				finalResult = FALSE;
			break;
			
		default:
			finalResult = FALSE;
		}
	}
	
	return finalResult;
}
