/*
 * gkacct.h
 *
 * Accounting modules for GNU Gatekeeper. Provides generic
 * support for accounting to the gatekeeper.
 *
 * Copyright (c) 2003, Quarcom FHU, Michal Zygmuntowicz
 *
 * This work is published under the GNU Public License (GPL)
 * see file COPYING for details.
 * We also explicitely grant the right to link this code
 * with the OpenH323 library.
 *
 * $Log: gkacct.h,v $
 * Revision 1.1.2.1  2003/06/19 15:36:04  zvision
 * Initial generic accounting support for GNU GK.
 *
 */
#ifndef __GKACCT_H_
#define __GKACCT_H_

#include <ptlib.h>
#include "RasTbl.h"

/** Module for logging accounting events
	generated by the gatekeeper.
*/
class GkAcctLogger : public PObject
{
	PCLASSINFO(GkAcctLogger,PObject)
public:
	/// Processing type for this module
	enum Control 
	{
		/// if cannot log an accounting event
		/// silently ignore the module and process remaining acct modules
		Optional, 
		/// if cannot log an accounting event do not allow futher 
		/// call processing (e.g. call should not be connected, etc.)
		/// always process remaining acct modules
		Required,
		/// if cannot log an accounting event
		/// stop processing and do not allow futher call progress
		/// (e.g. call should not be connected, etc.)
		/// if the event has been logged, do not process remaining acct modules
		Sufficient
	};

	/// status of the acct event processing
	enum Status 
	{
		Ok = 1,		/// acct even has been logged
		Fail = -1,	/// acct even has not been logged (failure)
		Next = 0	/// acct even has not been logged (undetermined status)
	};

	/// accounting event types
	enum AcctEvent 
	{
		AcctStart = 0x0001, /// start accounting a call
		AcctStop = 0x0002, /// stop accounting a call
		AcctUpdate = 0x0004, /// interim update
		AcctOn = 0x0008, /// accounting enabled (GK start)
		AcctOff = 0x0010, /// accounting disabled (GK stop)
		AcctAll = 0xffff,
		AcctNone = 0
	};
	
	/// Construct new accounting logger object.
	GkAcctLogger(
		/// config to be read for module settings
		PConfig& cfg, 
		/// module name (should be unique among different module types)
		const PString& moduleName = "default"
		);
		
	virtual ~GkAcctLogger();

	/** @return
		Name for this accounting module.
	*/
	PString GetName() const { return name; }

	/** @return
		Control flag determining processing behaviour for this module
		(optional,sufficient,required).
	*/
	Control GetControlFlag() const { return controlFlag; }

	/** @return
		Flags signalling which accounting events (see #AcctEvent enum#)
		should be logged. The flags are ORed.
	*/
	int GetEventMask() const { return eventMask; }
	
	/** Log accounting event.
	
		@return
		Status of this logging operation (see #Status enum#)
	*/
	virtual Status LogAcctEvent( 
		AcctEvent evt, /// accounting event to log
		callptr& call /// additional data for the event
		) 
	{
		return defaultStatus; 
	}

	/** Format time string suitable for accounting logs.
		Currently NTP format conversion occurs.
		
		@return
		Formatted time string.
	*/ 
	static PString AsString( 
		const PTime& tm /// time value to be formatted
		);

protected:
	/** @return
		Default status that should be returned when the request
		cannot be determined.
	*/
	Status GetDefaultStatus() const { return defaultStatus; }

private:
	/** Read an event mask (ORed #AccEvent enum# constants) 
		from the passed tokens.
	
		@return
		Resulting event mask.
	*/
	virtual int ReadEventMask( 
		const PStringArray& tokens /// event names (start from index 1)
		) const;
		
	GkAcctLogger(const GkAcctLogger &);
	GkAcctLogger & operator=(const GkAcctLogger &);
	 
private:
	/// processing behaviour (see #Control enum#)
	Control controlFlag;
	/// default processing status (see #Status enum#)
	Status defaultStatus;
	/// ORed #AcctEvent enum# constants - define which events are logged
	int eventMask;
	/// module name
	PString name;
};

/// Factory for instances of GkAcctLogger-derived classes
class GkAcctFactoryBase : public PObject
{
	PCLASSINFO(GkAcctFactoryBase,PObject)
protected:
	/** Consturct new factory and register its instance
		on the global list of all available acct logger
		factories.
	*/
	GkAcctFactoryBase( 
		/// name for the acct modules this factory creates
		const PString& moduleName = PString::Empty()
		);

public:	
	virtual ~GkAcctFactoryBase();
	
	/** Create new instance of the GkAcctLogger module.
	
		@return
		The new accounting logger module.
	*/
	virtual GkAcctLogger* CreateAcctLogger(
		PConfig& cfg /// settings for the new acct module
		) = 0;

	/** Compare two GkAcctLogger factories. Equality is defined
		as equality of factory names (case-sensitive). See #PObject#
		for more comments.
	*/
	virtual Comparison Compare(
		const PObject& obj
		) const;		

	/** @return
		A name for acct modules generates by this factory.
	*/
	PString GetName() const { return name; }

	/** Find factory object for specified GkAcctLogger modules.
	
		@return
		Factory object or NULL, if not matching factory found.
	*/
	static GkAcctFactoryBase* FindFactoryFor(
		const PString& moduleName /// name for acct modules
		);
		
private:
	/// name for acct modules generated by this factory
	PString name;

	PDECLARE_BASEARRAY(GkAcctFactoryArray,GkAcctFactoryBase*)
	};
	/// static list of registered logger factories
	static GkAcctFactoryArray* acctLoggerFactories;
};

/** Specialized factory template for easy declaration
	of new acct logger factories
*/
template<class _GkAcctLogger_> 
class GkAcctFactory : public GkAcctFactoryBase
{
	PCLASSINFO(GkAcctFactory,GkAcctFactoryBase)
public:
	GkAcctFactory(
		const PString& moduleName = PString::Empty()
		) 
		: 
		GkAcctFactoryBase(moduleName) 
	{
	}
	
	virtual GkAcctLogger* CreateAcctLogger(
		PConfig& cfg
		)
	{ 
		return new _GkAcctLogger_( cfg, GetName() ); 
	}
};

/** Build a list of accounting logger modules
	from the registered GkAcctFactories.
*/
PDECLARE_ARRAY(GkAcctLoggers,GkAcctLogger)
public:
	GkAcctLoggers(
		PConfig& cfg /// config file with module names and settings
		);
		
	virtual ~GkAcctLoggers();
	
	/** Log accounting event #evt# with all accounting logger modules
		on the list.
		
		@return
		TRUE if the event has been successfully logger. FALSE usually
		means to stop processing of further activities regarding the event
		(call should be terminated, rejected, etc.).
	*/
	virtual BOOL LogAcctEvent( 
		GkAcctLogger::AcctEvent evt, /// accounting event to log
		callptr& call /// additional data for the event
		);
};

#endif  /* __GKACCT_H_ */
