/*
 * radacct.cxx
 *
 * RADIUS protocol accounting logger module for GNU Gatekeeper. 
 *
 * Copyright (c) 2003, Quarcom FHU, Michal Zygmuntowicz
 *
 * This work is published under the GNU Public License (GPL)
 * see file COPYING for details.
 * We also explicitely grant the right to link this code
 * with the OpenH323 library.
 *
 */
#include <ptlib.h>
#include <h323pdu.h>
#include "h323util.h"
#include "gkacct.h"
#include "Toolkit.h"
#include "RasTbl.h"
#include "radacct.h"

// Settings for accoutning logger RADIUS module will be stored 
// inside [RadAcct] config section
#define RadAcctConfigSectionName "RadAcct"

// append RADIUS based accounting logger to the global list of loggers
GkAcctFactory<RadAcct> RAD_ACCT("RadAcct");

extern PString GetConferenceIDString( const H225_ConferenceIdentifier& id );

PMutex RadAcct::sessionMutex;
int RadAcct::sessionSeed = PRandom();
int RadAcct::sessionInc = 0;

RadAcct::RadAcct( 
	PConfig& cfg, 
	const PString& acctName 
	)
	:
	GkAcctLogger( cfg, acctName ),
	radiusServers( cfg.GetString(
			RadAcctConfigSectionName,"Servers",""
			).Tokenise( ";, \t", FALSE ) 
		),
	sharedSecret( cfg.GetString(
			RadAcctConfigSectionName,"SharedSecret",""
			)
		),
	acctPort( (WORD)(cfg.GetInteger(
			RadAcctConfigSectionName,"DefaultAcctPort"
			))
		),
	portBase( 1024 ),
	portMax( 65535 ),
	requestTimeout( cfg.GetInteger(
			RadAcctConfigSectionName,"RequestTimeout"
			)
		),
	idCacheTimeout( cfg.GetInteger(
			RadAcctConfigSectionName,"IdCacheTimeout"
			)
		),
	socketDeleteTimeout( cfg.GetInteger(
			RadAcctConfigSectionName,"SocketDeleteTimeout"
			)
		),
	numRequestRetransmissions( cfg.GetInteger(
			RadAcctConfigSectionName,"RequestRetransmissions"
			)
		),
	roundRobin( cfg.GetBoolean(
			RadAcctConfigSectionName,"RoundRobinServers"
			)
		),
	appendCiscoAttributes( cfg.GetBoolean(
			RadAcctConfigSectionName,"AppendCiscoAttributes"
			)
		),
	localInterface( cfg.GetString(
			RadAcctConfigSectionName, "LocalInterface", ""
			)
		),
	fixedUsername( cfg.GetString(
			RadAcctConfigSectionName, "FixedUsername", ""
			)
		),
	radiusClient( NULL )
{
	if( radiusServers.GetSize() < 1 )
	{
		PTRACE(1,"RADACCT\tCannot build "<<acctName<<" accounting logger"
			" - no RADIUS servers specified in the config"
			);
		return;
	}
	
	if( (!localInterface.IsEmpty()) 
		&& (!PIPSocket::IsLocalHost(localInterface)) )
	{
		PTRACE(1,"RADACCT\tConfigured local interface - "<<localInterface
			<<" - does not belong to this machine, assuming ip:*"
			);
		localInterface = PString::Empty();
	}

	/// build RADIUS client
	radiusClient = new RadiusClient( 
		radiusServers[0],
		(radiusServers.GetSize() > 1) ? radiusServers[1] : PString::Empty(),
		localInterface
		);

	/// if there were specified more than two RADIUS servers, append them
	for( int i = 2; i < radiusServers.GetSize(); i++ )
		radiusClient->AppendServer( radiusServers[i] );	
		
	radiusClient->SetSharedSecret( sharedSecret );
	radiusClient->SetRoundRobinServers( roundRobin );
		
	if( acctPort > 0 )
		radiusClient->SetAcctPort( acctPort );
		
	if( requestTimeout > 0 )
		radiusClient->SetRequestTimeout( requestTimeout );
	if( idCacheTimeout > 0 )
		radiusClient->SetIdCacheTimeout( idCacheTimeout );
	if( socketDeleteTimeout > 0 )
		radiusClient->SetSocketDeleteTimeout( socketDeleteTimeout );
	if( numRequestRetransmissions > 0 )
		radiusClient->SetRetryCount( numRequestRetransmissions );
	
	PStringArray s = cfg.GetString(
		RadAcctConfigSectionName,"RadiusPortRange",""
		).Tokenise( "-" );

	// parse port range		
	if( s.GetSize() >= 2 )
	{ 
		unsigned p1 = s[0].AsUnsigned();
		unsigned p2 = s[1].AsUnsigned();
	
		// swap if base is greater than max
		if( p2 < p1 )
		{
			const unsigned temp = p1;
			p1 = p2;
			p2 = temp;
		}
		
		if( p1 > 65535 )
			p1 = 65535;
		if( p2 > 65535 )
			p2 = 65535;
	
		if( (p1 > 0) && (p2 > 0) )
		{
			portBase = (WORD)p1;
			portMax = (WORD)p2;
		}
	}
	
	radiusClient->SetClientPortRange( portBase, portMax-portBase+1 );
}

RadAcct::~RadAcct()
{
	delete radiusClient;
}

GkAcctLogger::Status RadAcct::LogAcctEvent(
	GkAcctLogger::AcctEvent evt, 
	callptr& call
	)
{
	if( radiusClient == NULL )
	{
		PTRACE(1,"RADACCT\tAcct "<<((GetDefaultStatus() == Fail)?"failed":"skipped")
			<<" - null RADIUS client instance"
			);
		return GetDefaultStatus();
	}

	if( (evt & (AcctStart | AcctStop | AcctUpdate)) && (!call) )
	{
		PTRACE(2,"RADACCT\tAcct "<<((GetDefaultStatus() == Fail)?"failed":"skipped")
			<<" - missing call info"
			);
		return GetDefaultStatus();
	}
	
	if( (evt & RadAcctEvents) == 0 )
	{
		PTRACE(2,"RADACCT\tAcct "<<((GetDefaultStatus() == Fail)?"failed":"skipped")
			<<" - unknown event type "<<PString(PString::Unsigned,(long)evt,16)
			);
		return GetDefaultStatus();
	}
	
	// build RADIUS Accounting-Request
	RadiusPDU* pdu = radiusClient->BuildPDU();
	if( pdu == NULL )
	{
		PTRACE(2,"RADACCT\tAcct "<<((GetDefaultStatus()==Fail)?"failed":"skipped")
			<<" - could not build Accounting-Request PDU"
			);
		return GetDefaultStatus();
	}

	pdu->SetCode( RadiusPDU::AccountingRequest );

	*pdu += new RadiusAttr( RadiusAttr::AcctStatusType, 
		(evt & AcctStart) ? RadiusAttr::AcctStatus_Start
		: ((evt & AcctStop) ? RadiusAttr::AcctStatus_Stop
		: ((evt & AcctUpdate) ? RadiusAttr::AcctStatus_InterimUpdate
		: ((evt & AcctOn) ? RadiusAttr::AcctStatus_AccountingOn
		: ((evt & AcctOff) ? RadiusAttr::AcctStatus_AccountingOff : 0)
		))) );
			
	// Gk works as NAS point, so append NAS IP
	*pdu += new RadiusAttr( RadiusAttr::NasIpAddress, 
		localInterface.IsEmpty() 
			? Toolkit::Instance()->GetRouteTable()->GetLocalAddress()
			: PIPSocket::Address(localInterface)
		);
	*pdu += new RadiusAttr( RadiusAttr::NasIdentifier,
		Toolkit::Instance()->GKName()
		);
	*pdu += new RadiusAttr( RadiusAttr::NasPortType, RadiusAttr::NasPort_Virtual );
		
	if( evt & (AcctStart | AcctStop | AcctUpdate) )
	{
		*pdu += new RadiusAttr( RadiusAttr::ServiceType, RadiusAttr::ST_Login );
		
		PString sessionId = call->GetAcctSessionId();
		if( sessionId.IsEmpty() )
		{
			sessionId = GetSessionId();
			call->SetAcctSessionId(sessionId);
		}
		
		*pdu += new RadiusAttr( RadiusAttr::AcctSessionId, sessionId );


		PString srcInfo = call->GetSourceInfo();
		if( !srcInfo.IsEmpty() )
		{
			const PINDEX index = srcInfo.FindOneOf(":");
			if( index != P_MAX_INDEX )
				srcInfo = srcInfo.Left(index);
		}
	
		endptr callingEP = call->GetCallingParty();
		PString userName;
	
		if( callingEP && (callingEP->GetAliases().GetSize() > 0) )
			userName = GetBestAliasAddressString(
				callingEP->GetAliases(),
				H225_AliasAddress::e_h323_ID
				);
		else if( !srcInfo.IsEmpty() )
			userName = srcInfo;
		else if( call->GetCallingAddress() != NULL )
		{
			PIPSocket::Address addr;
			if( GetIPFromTransportAddr( *(call->GetCallingAddress()),addr) 
				&& addr.IsValid() )
				userName = addr.AsString();
		}
		
		if( !(userName.IsEmpty() && fixedUsername.IsEmpty()) )					
			*pdu += new RadiusAttr( RadiusAttr::UserName, 
				fixedUsername.IsEmpty() ? userName : fixedUsername 
				);
		else
			PTRACE(1,"RADACCT\tCould not determine User-Name for the call "<<call->GetCallNumber());
		
		if( (evt & AcctStart) == 0 )
		{
			PTimeInterval duration = 0;
	
			if( call->GetStartTime() != NULL )
				if( call->GetStopTime() != NULL )
				{
					if( *(call->GetStopTime()) < *(call->GetStartTime()) )
					{
						PTRACE(1,"RADACCT\tInvalid call timestamps - start: "
							<< *(call->GetStartTime()) << ", stop: "
							<< *(call->GetStopTime()) << ", call no "
							<< call->GetCallNumber()
							);
						duration = 0;
					}
					else
						duration = *(call->GetStopTime()) - *(call->GetStartTime());
				}
				else
				{
					if( (call->GetStartTime())->IsFuture() )
					{
						PTRACE(1,"RADACCT\tInvalid call timestamps - start: "
							<< *(call->GetStartTime()) << ", stop: "
							<< PTime() << ", call no "
							<< call->GetCallNumber()
							);
						duration = 0;
					}
					else
						duration = PTime() - *(call->GetStartTime());
				}
					
			*pdu += new RadiusAttr( 
				RadiusAttr::AcctSessionTime, 
				duration.GetSeconds() 
				);
		}
	
		PString callingStationId;
	
		callingStationId = call->GetCallingStationId();
		
		if( callingStationId.IsEmpty() && callingEP 
			&& (callingEP->GetAliases().GetSize() > 0) )
			callingStationId = GetBestAliasAddressString(
				callingEP->GetAliases(),
				H225_AliasAddress::e_dialedDigits,
				H225_AliasAddress::e_partyNumber,
				H225_AliasAddress::e_h323_ID
				);
					
		if( callingStationId.IsEmpty() )
			callingStationId = srcInfo;
			
		if( callingStationId.IsEmpty() && (call->GetCallingAddress() != NULL) )
		{
			PIPSocket::Address addr;
			if( GetIPFromTransportAddr( *(call->GetCallingAddress()),addr) 
				&& addr.IsValid() )
				callingStationId = addr.AsString();
		}
		
		if( !callingStationId.IsEmpty() )
			*pdu += new RadiusAttr( RadiusAttr::CallingStationId,
				callingStationId
				);
		else
			PTRACE(2,"RADACCT\tCould not determine Calling-Station-Id for the call "<<call->GetCallNumber());

		
		PString calledStationId;
								
		calledStationId = call->GetCalledStationId();
			
		if( calledStationId.IsEmpty() && !(call->GetDestInfo().IsEmpty()) )
		{
			const PString dest = call->GetDestInfo();
			const PINDEX index = dest.FindOneOf(":");
				
			calledStationId = (index == P_MAX_INDEX) ? dest : dest.Left(index);
		}
		
		if( calledStationId.IsEmpty() )
		{
			endptr calledEP = call->GetCalledParty();
			if( calledEP && (calledEP->GetAliases().GetSize() > 0) )
				calledStationId = GetBestAliasAddressString(
					calledEP->GetAliases(),
					H225_AliasAddress::e_dialedDigits,
					H225_AliasAddress::e_partyNumber,
					H225_AliasAddress::e_h323_ID
					);
		}
	
		if( calledStationId.IsEmpty() && (call->GetCalledAddress() != NULL) )
		{
			PIPSocket::Address addr;
			if( GetIPFromTransportAddr( *(call->GetCalledAddress()),addr) 
				&& addr.IsValid() )
				calledStationId = addr.AsString();
		}
		
		if( calledStationId.IsEmpty() )
			PTRACE(1,"GKACCT\tCould not determine Called-Station-Id for the call no "<<call->GetCallNumber());
		else
			*pdu += new RadiusAttr( RadiusAttr::CalledStationId, calledStationId );
		
		if( appendCiscoAttributes )
		{
			*pdu += new RadiusAttr(
				PString("h323-gw-id=") + Toolkit::Instance()->GKName(),
				CiscoVendorId,
				33
				);
			
			*pdu += new RadiusAttr(
				PString("h323-conf-id=") 
					+ GetConferenceIDString(call->GetConferenceIdentifier()),
				CiscoVendorId,
				24
				);
						
			*pdu += new RadiusAttr(
				PString("h323-call-origin=proxy"),
				CiscoVendorId,
				26
				);
				
			*pdu += new RadiusAttr(
				PString("h323-call-type=VoIP"),
				CiscoVendorId,
				27
				);
	
			if( evt & AcctStop )
			{
				call->SetSetupConnectDisconnect();

				PTime* tm = call->GetSetupTime();
		
				if( tm != NULL ) 					
					*pdu += new RadiusAttr(
						PString("h323-setup-time=") + AsString(*tm),
						CiscoVendorId,
						25
						);

				tm = call->GetConnectTime();
				if( tm != NULL )		
					*pdu += new RadiusAttr(
						PString("h323-connect-time=") + AsString(*tm),
						CiscoVendorId,
						28
						);
	
				tm = call->GetDisconnectTime();
				if( tm != NULL )
					*pdu += new RadiusAttr(
						PString("h323-disconnect-time=") + AsString(*tm),
						CiscoVendorId,
						29
						);
				
				*pdu += new RadiusAttr(
					PString("h323-disconnect-cause=0"),
					CiscoVendorId,
					30
					);
			}					
			
			if( call->GetCalledAddress() != NULL )
			{
				PIPSocket::Address addr;
				if( GetIPFromTransportAddr(*call->GetCalledAddress(),addr)
					&& addr.IsValid() )
					*pdu += new RadiusAttr(
						PString("h323-remote-address=") + addr.AsString(),
						CiscoVendorId,
						23
						);
			}
		}
	
		*pdu += new RadiusAttr( RadiusAttr::AcctDelayTime, 0 );
	}
		
	// send request and wait for response
	RadiusPDU* response = NULL;
	BOOL result = OnSendPDU(*pdu,(int)evt,call) 
		&& radiusClient->MakeRequest( *pdu, response ) 
		&& (response != NULL);
			
	delete pdu;
			
	if( !result )
	{
		delete response;
		PTRACE(3,"RADACCT\tAcct "<<((GetDefaultStatus()==Fail)?"failed":"undetermined")
			<<" - could not send Accounting-Request PDU (evt:"
			<<PString(PString::Unsigned,(long)evt,16)<<", call no:"
			<<(call?call->GetCallNumber():0)<<')'
			);
		return GetDefaultStatus();
	}
				
	// check if Access-Request has been accepted
	result = (response->GetCode() == RadiusPDU::AccountingResponse);
	if( result )
		result = OnReceivedPDU(*response,(int)evt,call);
						
	delete response;
	return result ? Ok : Fail;
}

BOOL RadAcct::OnSendPDU(
	RadiusPDU& pdu,
	int acctEventMask,
	callptr& call
	)
{
	return TRUE;
}

BOOL RadAcct::OnReceivedPDU(
	RadiusPDU& pdu,
	int acctEventMask,
	callptr& call
	)
{
	return TRUE;
}

PString RadAcct::GetSessionId()
{
	PWaitAndSignal lock( sessionMutex );
	
	return psprintf("%04x%04x",sessionSeed&0xffff,(sessionInc++)&0xffff);
}
