/*
 * radacct.h
 *
 * RADIUS protocol accounting module for GNU Gatekeeper. 
 *
 * Copyright (c) 2003, Quarcom FHU, Michal Zygmuntowicz
 *
 * This work is published under the GNU Public License (GPL)
 * see file COPYING for details.
 * We also explicitely grant the right to link this code
 * with the OpenH323 library.
 *
 */
#ifndef __RADACCT_H
#define __RADACCT_H

#include <ptlib.h>
#include "gkacct.h"
#include "radproto.h"


class RadAcct : public GkAcctLogger
{
	PCLASSINFO(RadAcct,GkAcctLogger)
public:
	enum Constants
	{
		/// events recognized by this module
		RadAcctEvents = AcctStart | AcctStop | AcctUpdate | AcctOn | AcctOff,
		CiscoVendorId = 9
	};
	
	/** Create GkAcctLogger for RADIUS protocol
	*/
	RadAcct( 
		/// accounting logger settings
		PConfig& cfg, 
		/// name from Gatekeeper::Acct section
		const PString& acctName 
		);
		
	/// Destroy the accounting logger
	virtual ~RadAcct();

	virtual Status LogAcctEvent(
		AcctEvent evt,
		callptr& call
		);

	/** Generate new unique RADIUS accounting session identifier.
	
		@return
		New session identifier string.
	*/
	static PString GetSessionId();

protected:		
	/** Called before Accounting-Request PDU is send.
		Can be used to introduce additional attributes etc.
		
		@return
		TRUE to proceed, FALSE to fail and not send this pdu.
	*/
	virtual BOOL OnSendPDU(
		RadiusPDU& pdu, /// PDU to be sent
		int acctEventMask, /// accounting event being processed
		callptr& call /// call associated with this request (if any)
		);

	virtual BOOL OnReceivedPDU(
		RadiusPDU& pdu, /// PDU received from RADIUS server
		int acctEventMask, /// accounting event being processed
		callptr& call /// call associated with this response (if any)
		);
		
private:
	/* No copy constructor allowed */
	RadAcct( const RadAcct& );
	/* No operator= allowed */
	RadAcct& operator=( const RadAcct& );
	
private:
	/// array of configured RADIUS server names
	PStringArray radiusServers;
	/// shared secret for gk client<->RADIUS server authorization
	PString sharedSecret;
	/// default port that will be used for sending RADIUS acct
	/// requests
	WORD acctPort;
	/// base port number for UDP client socket allocation
	WORD portBase;
	/// max port number for UDP client socket allocation
	WORD portMax;
	/// timeout (ms) for a single RADIUS request
	unsigned requestTimeout;
	/// timeout (ms) for RADIUS requests IDs to be unique
	unsigned idCacheTimeout;
	/// timeout (ms) for unused sockets to be deleted
	unsigned socketDeleteTimeout;
	/// how many times to transmit a single request (1==no retransmission)
	/// to a single RADIUS server
	unsigned numRequestRetransmissions;
	/// retransmission fashion: 
	/// 	FALSE - do #numRequestRetransmissions# for server A,
	///				then do #numRequestRetransmissions# for server B, etc.
	///		TRUE - transmit request to server A, then to server B, etc.
	///				the whole procedure repeat #numRequestRetransmissions# times
	BOOL roundRobin;
	/// if TRUE Cisco VSAs are appended to the RADIUS packets
	BOOL appendCiscoAttributes;
	/// Local interface RADIUS client should be bound to (multihomed hosts)
	PString localInterface;	
	/// Fixed value for User-Name attribute in outgoing requests
	PString fixedUsername;
	/// RADIUS protocol client class associated with this authenticator
	RadiusClient* radiusClient;
	
	static PMutex sessionMutex;
	static int sessionSeed;
	static int sessionInc;
};

#endif /* __RADACCT_H */
